/*
** Copyright (c) 2001  Hughes Technologies Pty Ltd.  All rights
** reserved.  
**
** Terms under which this software may be used or copied are
** provided in the  specific license associated with this product.
**
** Hughes Technologies disclaims all warranties with regard to this 
** software, including all implied warranties of merchantability and 
** fitness, in no event shall Hughes Technologies be liable for any 
** special, indirect or consequential damages or any damages whatsoever 
** resulting from loss of use, data or profits, whether in an action of 
** contract, negligence or other tortious action, arising out of or in 
** connection with the use or performance of this software.
**
**
** $Id: symtab.c,v 1.16 2002/06/04 06:37:24 bambi Exp $
**
*/

/*
** Module	: 
** Purpose	: 
** Exports	: 
** Depends Upon	: 
*/



/*
** Notes
**
** These are the symbol table handling routines.  The symtab is
** constructed in a hash table.  As an entry is created it is assigned
** a unique ID number which is used to reference the symbol during
** script execution (i.e. all symbol references in the pcode will be
** via the numeric ID not the textual name). This improves execution
** speed as it does not require repeated string comparisons to find
** symbols.
**
** Naturally, we need to access the symbols by name during the parse.
** As a result there is a stock standard X*Y hash table scan access
** method available via the name lookup function.
*/ 

/**************************************************************************
** STANDARD INCLUDES
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "config/config.h"
#include "ember.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif


/**************************************************************************
** MODULE SPECIFIC INCLUDES
**************************************************************************/

#include "parser.h"

/**************************************************************************
** GLOBAL VARIABLES
**************************************************************************/


/**************************************************************************
** PRIVATE ROUTINES
**************************************************************************/

static void debugTrap()
{
}

static int _lookupLiteralByValue(script, value, id)
	ember	*script;
	e_value	*value;
	int	*id;
{
	REG	int	count;
	REG	e_stab	*curSym;
	int	res;

	count=0;
	while(count < EMBER_HASH_LEN)
	{
		curSym = script->literals[count];
		while(curSym)
		{
			if (curSym->value.type == value->type)
			{
				if (eTypeCompareValues(script,value->type,
					value, &curSym->value, &res) < 0)
				{
					return(-1);
				}
				if (res == 1)
				{
					*id = curSym->id;
					return(0);
				}
			}
			curSym = curSym->next;
		}
		count++;
	}
	*id = 0;
	return(0);
}


/**************************************************************************
** PUBLIC ROUTINES
**************************************************************************/



/**************************************************************************
** Variable Map Routines
** =====================
**
** The variable map is a hash table that contains mappings from a
** variable name to a numeric symbol ID.  There is a variable map
** contained in the definition of a function (local variables) and
** also in the top scope (main scope variables).
**
** The map is only used during code generation.  All variable references
** at run-time use the numeric IDs for symbol table lookup.  The map
** not only provides for the name->id translations, but it also defines
** the visibility scope of a symbol
*/

e_map *eSymFindVariableMap(script,map, name)
	ember	*script;
	e_map	*map[];
	char	*name;
{
	REG	e_map *curMap;
	int	hash;

	/*
	** Dodge a leading var designator
	*/
	if (*name == '$' || *name == '@')
		name++;
	hash = eGenHash(name);
	curMap = map[hash];
	while(curMap)
	{
		if (*name == *curMap->name &&
			strcmp(name, curMap->name) == 0)
		{
			return(curMap);
		}
		curMap = curMap->next;
	}
	sprintf(script->errorMsg,"Can't find variable map for '%s'",name);
	if (strcmp(name,"ERRMSG") == 0)
		debugTrap();
	return(NULL);
}


int eSymAddVariableMap(script, name, id)
	ember	*script;
	char	*name;
	int	id;
{
	e_map	*curMap,
		**base;
	int	hash;
	char	*namePtr;


	/*
	** Work out where to put this symbol.  If it's a normal '$'
	** variable then put it in the local scope.  If it's a '@'
	** variable then it references something in the main scope
	*/
	if (*name == '@')
	{
		/* Global / Main variable */
		base = script->main.variables;
	}
	else
	{
		/* Local variable */
		base = script->info.curFunction->variables;
	}

	/*
	** Write the map entry.  Note: We drop the '$' or '@' from the
	** variable and just enter the name
	*/
	if (*name == '$' || *name == '@')
		namePtr = name + 1;
	else
		namePtr = name;
	
	hash = eGenHash(namePtr);
	curMap = base[hash];
	if (curMap == NULL)
	{
		base[hash] = (e_map *)eMemMalloc(script, sizeof(e_map));
		curMap = base[hash];
	}
	else
	{
		while(curMap->next)
			curMap = curMap->next;
		curMap->next = (e_map *)eMemMalloc(script, sizeof(e_map));
		curMap = curMap->next;
	}
	if (curMap == NULL)
		return(-1);
	strcpy(curMap->name, namePtr);
	curMap->id = id;
	curMap->next = NULL;
	return(0);
}





/**************************************************************************
** Symbol Table Routines
** =====================
**
** There is a single symbol table in which all variables, both locals
** and globals, are defined and manipulated.  The symbols are referenced
** by numeric ID which has the effect of hiding the variable name.  This
** allows multiple symbols from different scopes to be defined with the
** same name in the same table (see variable map above)
*/

void eSymDupValue(script, source, dest)
	ember	*script;
	e_value	*source,
		*dest;
{
	e_array	*curElement;
	e_assoc	*curAssoc;
	int	count;

	dest->typePtr = source->typePtr;
	dest->type = source->type;
	if (source->initialised == EMBER_FALSE)
	{
		dest->initialised = EMBER_FALSE;
		return;
	}
	switch(eTypeGetBaseType(source->type))
	{
		case EMBER_TYPE_INT:
			dest->value.intVal = source->value.intVal;
			break;

		case EMBER_TYPE_UINT:
			dest->value.uintVal = source->value.uintVal;
			break;

		case EMBER_TYPE_REAL:
			dest->value.realVal = source->value.realVal;
			break;

		case EMBER_TYPE_TEXT:
			dest->value.textVal = 
				strdup(source->value.textVal);
			break;
		case EMBER_TYPE_ARRAY:
			count = 0;
			curElement = (e_array*)source->value.arrayVal;
			while(curElement)
			{
				eSymSetArrayElement(script, dest,
					&curElement->value, count);
				count++;
				curElement = curElement->next;
			}
			break;
		case EMBER_TYPE_ASSOC:
			curAssoc = (e_assoc*)source->value.assocVal;
			while(curAssoc)
			{
				eSymSetAssocElement(script, dest,
					&curAssoc->value, curAssoc->index);
				curAssoc = curAssoc->next;
			}
			break;
	}
	dest->initialised = EMBER_TRUE;
}


int eSymExpandLiteral(script, source, dest)
	ember	*script;
	e_value	*source,
		*dest;
{
	char	*cp;

	dest->typePtr = source->typePtr;
	dest->type = source->type;
	dest->initialised = EMBER_TRUE;
	switch(eTypeGetBaseType(source->type))
	{
		case EMBER_TYPE_INT:
			dest->value.intVal = source->value.intVal;
			break;

		case EMBER_TYPE_UINT:
			dest->value.uintVal = source->value.uintVal;
			break;

		case EMBER_TYPE_REAL:
			dest->value.realVal = source->value.realVal;
			break;

		case EMBER_TYPE_TEXT:
			cp = eRunExpandText(script,source->value.textVal);
			if (!cp)
				return(-1);
			dest->value.textVal = strdup(cp);
			break;
	}
	return(0);
}




e_stab *eSymGetVariableById(script, id)
	ember	*script;
	int	id;
{
	REG	e_stab	*curSym;

	curSym = script->variables[id % EMBER_HASH_LEN];
	while(curSym)
	{
		if (curSym->id == id)
		{
			return(curSym);
		}
		curSym = curSym->next;
	}
	return(NULL);
}



e_stab *eSymGetVariableByName(script, name)
	ember	*script;
	char	*name ;
{
	e_map	**mapBase,
		*map;
	static	e_stab	symbol;
	char	*envVar;
	int	state;	/* 1 = local, 2 = global */

	/*
	** Find the scope we need to search in
	*/
	if (*name == '@')
		state = 2;
	else
		state = 1;
	if (*name == '$' || *name == '@')
		name++;

	while (state < 3)
	{
		if (state == 2)
		{
			mapBase = script->main.variables;
		}
		if (state == 1)
		{
			mapBase = script->info.curFunction->variables;
		}

		/*
		** Lookup the map entry and return the symbol if we find it
		*/
		map = eSymFindVariableMap(script, mapBase, name);

		if (map)
			break;
		state++;
	}
	if (map)
	{
		eUtilResetError(script);
		return(eSymGetVariableById(script, map->id));
	}

	/*
	** Could be environment variable! 
	*/
	envVar = getenv(name);
	if (envVar == NULL)
		return(NULL);
	eUtilResetError(script);
	eSymSetTextValue(script,&(symbol.value), envVar);
	return(&symbol);
}


e_stab *eSymCreateVariable(script, name)
	ember	*script;
	char	*name;
{
	REG 	e_stab	*curSym;
	int 	hash;
	char	*namePtr;

	/*
	** Is the name within limits?
	*/
	if (strlen(name + 1) > EMBER_NAME_LEN)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Variable name too long (max is %d characters)",
			EMBER_NAME_LEN);
		return(NULL);
	}


	/*
	** Create the new symbol
	*/
	script->info.lastSymID++;
	hash = script->info.lastSymID % EMBER_HASH_LEN;
	curSym = script->variables[hash];
	if (curSym == NULL)
	{
		script->variables[hash] = (e_stab*)eMemMalloc(script,
			sizeof(e_stab));
		curSym = script->variables[hash];
	}
	else
	{
		while(curSym->next)
		{
			curSym = curSym->next;
		}
		curSym->next = (e_stab*)eMemMalloc(script,sizeof(e_stab));
		curSym = curSym->next;
	}
	if (curSym == NULL)
		return(NULL);
	bzero(curSym,sizeof(e_stab));


	/*
	** Store the details
	*/
	if (*name == '$' || *name == '@')
		namePtr = name + 1;
	else
		namePtr = name;
	strcpy(curSym->name,namePtr);
	curSym->id = script->info.lastSymID;
	script->info.lastSymID = script->info.lastSymID;
	eSymAddVariableMap(script, name, script->info.lastSymID);
	return(curSym);
}


char *eSymGetVariableValue(script, name)
	ember	*script;
	char	*name;
{
	e_stab	*curSym;
	char	*res;

	curSym = eSymGetVariableByName(script, name);
	if (curSym == NULL)
		return(NULL);
	res = (*curSym->value.typePtr->unpack) (&curSym->value);
	return(res);
}


void eSymInitArray(script, value)
	ember	*script;
	e_value	*value;
{
	value->type = EMBER_TYPE_ARRAY;
	value->initialised = EMBER_TRUE;
	value->value.arrayVal = NULL;
}


e_value *eSymGetArrayElement(script, value, element)
	ember	*script;
	e_value	*value;
	int	element;
{
	e_array	*curElement;
	int	count;

	if (value->type != EMBER_TYPE_ARRAY)
		return(NULL);
	count = 0;
	curElement = (e_array*)value->value.arrayVal;
	while(count < element && curElement)
	{
		curElement = curElement->next;
		count++;
	}
	if (count < element)
		return(NULL);
	return(&curElement->value);
}


int eSymGetNumArrayElements(script, value)
	ember	*script;
	e_value	*value;
{
	int	count;
	e_array	*curElement;

	count = 0;
	curElement = value->value.arrayVal;
	while(curElement)
	{
		count++;
		curElement = curElement->next;
	}
	return(count);
}


e_value *eSymGetAssocElement(script, value, index)
	ember	*script;
	e_value	*value;
	char	*index;
{
	e_assoc	*curElement;

	if (value->type != EMBER_TYPE_ASSOC)
		return(NULL);
	curElement = (e_assoc*)value->value.assocVal;
	while(curElement)
	{
		if (strcmp(curElement->index, index) == 0)
			break;
		curElement = curElement->next;
	}
	if (curElement)
		return(&curElement->value);
	return(NULL);
}

int eSymGetNumAssocElements(script, value)
	ember	*script;
	e_value	*value;
{
	int	count;
	e_assoc	*curElement;

	count = 0;
	curElement = value->value.assocVal;
	while(curElement)
	{
		count++;
		curElement = curElement->next;
	}
	return(count);
}


/**************************************************************************
** Value Creation Routines
** =======================
*/

int eSymSetArrayElement(script, varPtr, valPtr, element)
	ember	*script;
	e_value	*varPtr,
		*valPtr;
	int	element;
{
	e_array	*curElement,
		*prevElement;
	int	count;

	if (element < 0)
	{
		strcpy(script->errorMsg,
			"Negative array dereferences are not valid");
		return(-1);
	}
	varPtr->type = EMBER_TYPE_ARRAY;
	if (varPtr->initialised == EMBER_FALSE)
		varPtr->value.arrayVal = NULL;

	if (varPtr->value.arrayVal == NULL)
	{
		varPtr->value.arrayVal = (void*)malloc(sizeof(e_array));
		curElement = (e_array *)varPtr->value.arrayVal;
		curElement->value.initialised = 0;
		curElement->value.value.intVal = 0;
		curElement->next = NULL;
		
	}
	varPtr->initialised = EMBER_TRUE;
	count = 0;
	curElement = (e_array *)varPtr->value.arrayVal;
	while(count < element)
	{
		prevElement = curElement;
		curElement = curElement->next;
		if (!curElement)
		{
			prevElement->next = (e_array*)malloc(sizeof(e_array));
			curElement = prevElement->next;
			curElement->value.initialised = 0;
			curElement->next = NULL;
		}
		count++;
	}
	if (curElement->value.initialised)
		eSymFreeValue(script, &curElement->value);
	eSymDupValue(script, valPtr, &curElement->value);
	return(0);
}



int eSymSetAssocElement(script, varPtr, valPtr, index)
	ember	*script;
	e_value	*varPtr,
		*valPtr;
	char	*index;
{
	e_assoc	*curElement,
		*newElement,
		*prevElement;

	varPtr->type = EMBER_TYPE_ASSOC;

	curElement = varPtr->value.assocVal;
	prevElement = newElement = NULL;
	while(curElement)
	{
		if (strcmp(curElement->index,index) == 0)
			break;
		prevElement = curElement;
		curElement = curElement->next;
	}
	if (curElement)
	{
		/* Reset existing assoc entry */
		eSymFreeValue(script, &curElement->value);
		eSymDupValue(script, valPtr, &curElement->value);
	}
	else
	{
		/* Create a new assoc entry */
		newElement = (void*)malloc(sizeof(e_assoc));
		bzero(newElement, sizeof(e_assoc));
		if (prevElement)
			prevElement->next = newElement;
		else
			varPtr->value.assocVal = newElement;
		newElement->index = (char *)strdup(index);
		eSymDupValue(script, valPtr, &newElement->value);
	}
	return(0);
}



int eSymSetIntValue(script, valPtr, value)
	ember	*script;
	e_value	*valPtr;
	int	value;
{
	if (valPtr->initialised == EMBER_TRUE)
		eSymFreeValue(script, valPtr);
	valPtr->value.intVal = value;
	valPtr->type = EMBER_TYPE_INT;
	valPtr->typePtr = eTypeGetType(EMBER_TYPE_INT);
	valPtr->length = 4;
	valPtr->initialised = EMBER_TRUE;
	return(0);
}


int eSymSetUintValue(script, valPtr, value)
	ember	*script;
	e_value	*valPtr;
	u_int	value;
{
	if (valPtr->initialised == EMBER_TRUE)
		eSymFreeValue(script, valPtr);
	valPtr->value.uintVal = value;
	valPtr->type = EMBER_TYPE_UINT;
	valPtr->typePtr = eTypeGetType(EMBER_TYPE_UINT);
	valPtr->length = 4;
	valPtr->initialised = EMBER_TRUE;
	return(0);
}


int eSymSetRealValue(script, valPtr, value)
	ember	*script;
	e_value	*valPtr;
	double	value;
{
	if (valPtr->initialised == EMBER_TRUE)
		eSymFreeValue(script, valPtr);
	valPtr->value.realVal = value;
	valPtr->type = EMBER_TYPE_REAL;
	valPtr->typePtr = eTypeGetType(EMBER_TYPE_REAL);
	valPtr->length = 8;
	valPtr->initialised = EMBER_TRUE;
	return(0);
}


int eSymSetTextValue(script, valPtr, value)
	ember	*script;
	e_value	*valPtr;
	char	*value;
{
	if (valPtr->initialised == EMBER_TRUE)
		eSymFreeValue(script, valPtr);
	valPtr->value.textVal = eMemStrdup(script,value);
	if (valPtr->value.textVal == NULL)
		return(-1);
	valPtr->type = EMBER_TYPE_TEXT;
	valPtr->typePtr = eTypeGetType(EMBER_TYPE_TEXT);
	valPtr->length = strlen(value) + 1;
	valPtr->initialised = EMBER_TRUE;
	return(0);
}



/**************************************************************************
** Literal Value Routines
** ======================
**
** Literal values are represented as symbols without a name is a
** separate symbol table.  The symbol has no name (nor var map).  The
** mapping to a literal's symbol ID is based on the literal's type and
** value.  We ensure that only one stab entry is created per distinct
** literal value used in the script
*/

int eSymCreateLiteralFromValue(script,value, id)
	ember	*script;
	e_value	*value;
	int	*id;
{
	int	hash;
	e_stab	*curSym;

	/*
	** Find an existing matching literal if there is one
	*/
	if(_lookupLiteralByValue(script, value, id) < 0)
		return(-1);
	if (*id > 0)
		return(1);

	/*
	** Hmmm, couldn't find a matching literal.  Must make a new
	** one.
	*/

	script->info.lastLitID++;
	hash = script->info.lastLitID % EMBER_HASH_LEN;
	curSym = script->literals[hash];
	if (curSym == NULL)
	{
		script->literals[hash] = (e_stab*)eMemMalloc(script,
			sizeof(e_stab));
		curSym = script->literals[hash];
	}
	else
	{
		while(curSym->next)
		{
			curSym = curSym->next;
		}
		curSym->next = (e_stab*)eMemMalloc(script,sizeof(e_stab));
		curSym = curSym->next;
	}
	if (curSym == NULL)
		return(-1);
	bzero(curSym,sizeof(e_stab));

	curSym->id = script->info.lastLitID;
	curSym->value.initialised = EMBER_TRUE;
	bcopy(value, &curSym->value, sizeof(e_value));
	*id = curSym->id;
	return(0);
}


int eSymCreateLiteral(script,type,textRep, id)
	ember	*script;
	int	type;
	char	*textRep;
	int	*id;
{
	e_value	value;
	e_type	*typePtr;
	int	wasHtml,
		res;

	/*
	** If the text rep is in quoted string format we need to
	** blow away the quotes
	*/
	wasHtml = 0;
	if (type == HTML)
	{
		wasHtml = 1;
		type = EMBER_TYPE_TEXT;
	}
	typePtr = eTypeGetType(type);
	if (eTypeGetBaseType(type) == EMBER_TYPE_TEXT)
	{
		if (wasHtml)
		{
			(*typePtr->pack)(textRep, &value);
		}
		else
		{
			/* blow away quotes if we need to */
			if (*textRep == '"')
			{
				(*typePtr->pack)(textRep + 1, &value);
				value.value.textVal[
					strlen(value.value.textVal)-1]=0;
				value.length--;
			}
			else
			{
				(*typePtr->pack)(textRep, &value);
			}
		}
	}
	else
	{
		(*typePtr->pack)(textRep, &value);
	}
	value.initialised = EMBER_TRUE;
	value.type = type;
	value.typePtr = typePtr;

	res = eSymCreateLiteralFromValue(script,&value,id);
	if (res < 0 || res == 1)
	{
		eSymFreeValue(script, &value);
	}
	return(res);
}


e_stab *eSymGetLiteral(script, id)
	ember	*script;
	int	id;
{
	REG 	e_stab *curSym;

	curSym = script->literals[id % EMBER_HASH_LEN];
	while(curSym)
	{
		if (curSym->id == id)
		{
			return(curSym);
		}
		curSym = curSym->next;
	}
	return(NULL);
}





/**************************************************************************
** Function Routines
** =================
*/

e_funct *eSymGetFunctionById(script, id)
	ember	*script;
	int	id;
{
	REG	e_funct	*curFunct;

	curFunct = script->scriptFuncts[id % EMBER_HASH_LEN];
	while(curFunct)
	{
		if (curFunct->id == id)
		{
			return(curFunct);
		}
		curFunct = curFunct->next;
	}
	return(NULL);
}



e_funct *eSymGetFunctionByName(script, name)
	ember	*script;
	char	*name;
{
	REG	int	count;
	REG	e_funct	*curFunct;
	int	state;
	e_funct	**functions;

	state = 0;
	functions = script->scriptFuncts;
	while(state < 2)
	{
		count=0;
		while(count < EMBER_HASH_LEN)
		{
			curFunct = functions[count];
			while(curFunct)
			{
				if (*curFunct->name == *name &&
			    	strcmp(curFunct->name,name) == 0)
				{
					return(curFunct);
				}
				curFunct = curFunct->next;
			}
			count++;
		}
		state++;
		if (state == 1)
		{
			functions = script->moduleFuncts;
		}
	}
	snprintf(script->errorMsg,EMBER_ERR_LEN, "Unknown function '%s'",name);
	return(NULL);
}



int eSymCreateFunction(script, name)
	ember	*script;
	char	*name;
{
	int 	hash;
	REG 	e_funct	*curFunct;


	/*
	** Is the name within limits?
	*/
	if (strlen(name) > EMBER_NAME_LEN)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Function name too long (max is %d characters)",
			EMBER_NAME_LEN);
		return(-1);
	}


	/*
	** Check for duplicates.  If we find a similarly named script
	** that is a module function then we can still create this
	** function and override the module function.
	*/
	curFunct = eSymGetFunctionByName(script, name);
	if (curFunct && curFunct->type == EMBER_SCRIPT_FUNCTION)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Duplicate function name '%s'", name);
		return(-1);
	}

	*script->errorMsg = 0;

	/*
	** Create the function table entry
	*/
	script->info.lastFunctID++;
	hash = script->info.lastFunctID % EMBER_HASH_LEN;
	curFunct = script->scriptFuncts[hash];
	if (curFunct == NULL)
	{
		script->scriptFuncts[hash] = (e_funct*)eMemMalloc(script,
			sizeof(e_funct));
		curFunct = script->scriptFuncts[hash];
	}
	else
	{
		while(curFunct->next)
		{
			curFunct = curFunct->next;
		}
		curFunct->next = (e_funct*)eMemMalloc(script,sizeof(e_funct));
		curFunct = curFunct->next;
	}
	if (curFunct == NULL)
		return(-1);
	bzero(curFunct,sizeof(e_funct));

	strcpy(curFunct->name,name);
	curFunct->id = script->info.lastFunctID;
	curFunct->type = EMBER_SCRIPT_FUNCTION;
	script->info.curFunction = curFunct;
	return(0);
}


int eSymAddFunctionArg(script, arg, type)
	ember	*script;
	char	*arg;
	int	type;
{
	e_arg	*curArg,
		*tailArg;
	e_stab	*curSym;

	if (*arg == '$')
		arg++;
	curArg = (e_arg*)eMemMalloc(script,sizeof(e_arg));
	if (!curArg)
		return(-1);
	curSym = eSymCreateVariable(script,arg);
	strcpy(curArg->name, arg);
	curArg->type = type;
	curArg->id = curSym->id;
	curArg->ptr = curSym;
	curArg->next = NULL;
	if(script->info.curFunction->args == NULL)
	{
		script->info.curFunction->args = curArg;
	}
	else
	{
		tailArg = script->info.curFunction->args;
		while(tailArg->next)
		{
			tailArg = tailArg->next;
		}
		tailArg->next = curArg;
	}
	script->info.curFunction->numArgs++;
	return(0);
}


int eSymAddExternalFunction(script, efunct)
	ember	*script;
	e_extern *efunct;
{
	int 	hash;
	char	*name;
	REG 	e_funct	*curFunct;


	/*
	** Is the name within limits?
	*/
	name = efunct->name;
	if (strlen(name) > EMBER_NAME_LEN)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Function name too long (max is %d characters)",
			EMBER_NAME_LEN);
		return(-1);
	}


	/*
	** Check for duplicates.  If we find a similarly named script
	** function then don't add this module function (it has been
	** overridden).  If it's a module function then yell and scream
	** about duplicate names.
	*/
	curFunct = eSymGetFunctionByName(script, name);
	if (curFunct && curFunct->type == EMBER_SCRIPT_FUNCTION)
	{
		return(0);
	}
	if (curFunct && curFunct->type == EMBER_MODULE_FUNCTION)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Duplicate module function name '%s'", name);
		return(-1);
	}
	*script->errorMsg = 0;

	/*
	** Create the function table entry
	*/
	script->info.lastFunctID++;
	hash = script->info.lastFunctID % EMBER_HASH_LEN;
	curFunct = script->moduleFuncts[hash];
	if (curFunct == NULL)
	{
		script->moduleFuncts[hash] = (e_funct*)eMemMalloc(script,
			sizeof(e_funct));
		curFunct = script->moduleFuncts[hash];
	}
	else
	{
		while(curFunct->next)
		{
			curFunct = curFunct->next;
		}
		curFunct->next = (e_funct*)eMemMalloc(script,sizeof(e_funct));
		curFunct = curFunct->next;
	}
	if (curFunct == NULL)
		return(-1);
	bzero(curFunct,sizeof(e_funct));

	strcpy(curFunct->name,name);
	curFunct->id = script->info.lastFunctID;
	curFunct->type = EMBER_MODULE_FUNCTION;
	curFunct->externPtr = efunct;
	return(0);
}



/**************************************************************************
** Macro Handling Routines
** =======================
*/

int eSymCreateMacroFromValue(script, name, value)
	ember	*script;
	char	*name;
	e_value	*value;
{
	int	id,
		hash;
	e_stab	*tmp;
	e_map	*curMap;

	curMap = eSymFindVariableMap(script, script->macros, name);
	if (curMap)
	{
		snprintf(script->errorMsg, EMBER_ERR_LEN,
			"Macro '%s' already defined",
			name);
		return(-1);
	}
	eSymCreateLiteralFromValue(script,value, &id);
	tmp = eSymGetLiteral(script, id);

	hash = eGenHash(name);
	curMap = script->macros[hash];
	if (curMap == NULL)
	{
		script->macros[hash]=(e_map *)eMemMalloc(script,sizeof(e_map));
		curMap = script->macros[hash];
	}
	else
	{
		while(curMap->next)
			curMap = curMap->next;
		curMap->next = (e_map *)eMemMalloc(script, sizeof(e_map));
		curMap = curMap->next;
	}
	if (curMap == NULL)
	{
		return(-1);
	}
	strcpy(curMap->name, name);
	curMap->id = id;
	curMap->next = NULL;
	return(0);
}


int eSymCreateTextMacro(script, name, textVal)
	ember	*script;
	char	*name;
	char	*textVal;
{
	e_value	value;

	if (eSymSetTextValue(script, &value, textVal) < 0)
	{
		return(-1);
	}
	return(eSymCreateMacroFromValue(script, name, &value));
}


int eSymCreateIntMacro(script, name, intVal)
	ember	*script;
	char	*name;
	int	intVal;
{
	e_value	value;

	if (eSymSetIntValue(script, &value, intVal) < 0)
	{
		return(-1);
	}
	return(eSymCreateMacroFromValue(script, name, &value));
}


int eSymCreateRealMacro(script, name, realVal)
	ember	*script;
	char	*name;
	double	realVal;
{
	e_value	value;

	if (eSymSetRealValue(script, &value, realVal) < 0)
	{
		return(-1);
	}
	return(eSymCreateMacroFromValue(script, name, &value));
}



/**************************************************************************
** Utility Routines
** ================
*/

void eSymDumpSymbolMap(script, base)
	ember	*script;
	e_map	*base[];
{
	int	count=0;
	e_map	*curMap;
	e_stab	*curSym;
	char	outBuf[100];

	printf("    Name          ID     Initialised\n");
	printf("  ----------------------------------\n");

	while(count < EMBER_HASH_LEN)
	{
		curMap = base[count];
		while(curMap)
		{
			curSym = eSymGetVariableById(script, curMap->id);
			printf("   %-14.14s %-4d      %c",curSym->name, 
				curSym->id, curSym->value.initialised?'Y':'N');
			if(curSym->value.initialised &&
				curSym->value.type != EMBER_TYPE_ARRAY)
			{
				snprintf(outBuf, 30, "  '%s'",
				    (*curSym->value.typePtr->unpack)
					(&curSym->value));
				printf(outBuf);
			}
			printf("\n");
			curMap = curMap->next;
		}
		count++;
	}
}



void eSymDumpSymbols(script, base)
	ember	*script;
	e_stab	*base[];
{
	int	count=0;
	char	*type,
		*cp;
	e_stab	*curSym;

	printf("    ID   Type        Value \n");
	printf("  ---------------------------------------------------\n");

	while(count < EMBER_HASH_LEN)
	{
		curSym = base[count];
		while(curSym)
		{
			type = "unknown";
			if (curSym->value.typePtr)
				type = curSym->value.typePtr->name;
			if (curSym->value.type == EMBER_TYPE_ARRAY)
				type = "array";
			printf("    %-4d %-8s", curSym->id, type);

			if (curSym->value.initialised)
			{
				cp = (*curSym->value.typePtr->unpack)
					(&curSym->value);
				fflush(stdout);
				write(1,cp, strlen(cp)>30?30:strlen(cp));
				/*
				snprintf(outBuf, 30, "%s",
				    (*curSym->value.typePtr->unpack)
				printf(outBuf);
				*/
			}
			printf("\n");
			curSym = curSym->next;
		}
		count++;
	}
}



void eSymFreeValue(script, value)
	ember	*script;
	e_value	*value;
{
	e_array	*curElement,
		*prevElement;
	e_assoc *curAssoc,
		*prevAssoc;

	if (value->initialised == EMBER_FALSE)
		return;
	switch(value->type)
	{
	    case EMBER_TYPE_TEXT:
		free(value->value.textVal);
		break;

	    case EMBER_TYPE_ARRAY:
		curElement = (e_array*)value->value.arrayVal;
		while(curElement)
		{
			prevElement = curElement;
			curElement = curElement->next;
			prevElement->next = NULL;
			eSymFreeValue(script,&prevElement->value);
			free(prevElement);
		}
		value->value.arrayVal = NULL;
		break;

	    case EMBER_TYPE_ASSOC:
		curAssoc = (e_assoc *)value->value.assocVal;
		while(curAssoc)
		{
			prevAssoc = curAssoc;
			curAssoc = curAssoc->next;
			eSymFreeValue(script,&prevAssoc->value);
			free(prevAssoc->index);
			free(prevAssoc);
		}
		break;
	}
	value->initialised = EMBER_FALSE;
	value->type = 0;
	return;
}



void eSymFreeSymbolMap(script, base)
	ember	*script;
	e_map	*base[];
{
	int	count=0;
	e_map	*curMap,
		*prevMap;

	while(count < EMBER_HASH_LEN)
	{
		curMap = base[count];
		while(curMap)
		{
			prevMap = curMap;
			curMap = curMap->next;
			free(prevMap);
		}
		count++;
	}
}



void eSymFreeSymbols(script, base)
	ember	*script;
	e_stab	*base[];
{
	int	count=0;
	e_stab	*curSym,
		*prevSym;

	while(count < EMBER_HASH_LEN)
	{
		curSym = base[count];
		while(curSym)
		{
			eSymFreeValue(script, &curSym->value);
			prevSym = curSym;
			curSym = curSym->next;
			free(prevSym);
		}
		count++;
	}
}

void eSymFreeArgList(script, funct)
	ember	*script;
	e_funct	*funct;
{
	e_arg	*curArg,
		*prevArg;

	curArg = funct->args;
	while(curArg)
	{
		prevArg = curArg;
		curArg = curArg->next;
		free(prevArg);
	}
}


